/*
 * @author     Ekawat (ice) Homsirikamol
 * @version    1.0
 * @copyright  Copyright (c) 2017 Cryptographic Engineering Research Group
 *             ECE Department, George Mason University Fairfax, VA, U.S.A.
 *             All rights Reserved.
 * @license    This project is released under the GNU Public License.
 *             The license and distribution terms for this file may be
 *             found in the file LICENSE in this distribution or at
 *             http://www.gnu.org/licenses/gpl-3.0.txt
 * @note       This implementation is derived in part from the reference implementation
 *             included in SUPERCOP-20171020, with the list of co-authors listed in ref/designers
 *             
 *             This is publicly available encryption source code that falls
 *             under the License Exception TSU (Technology and software-
 *             unrestricted)
 */

#if !__SYNTHESIS__
#include <iostream>
#endif

#include "ap_int.h"
#include "hls_stream.h"
#include "hls_interface.h"
using namespace hls;

#if !__SYNTHESIS__
#define DBG 1
#define DBG_INPUT 0
#endif
#define SYNTHESIS_FAIL 0

#define PBS 48
#define nn 48
#define mm 2

#define ROTL(X, n) ((X << n) | (X >> (PBS-n)))

enum fsms_t {S_INIT, 
    S_AD, S_AD_FULL,
    S_STORE_MSG, S_MSG, S_MSG_FULL,
    S_TAG, S_TAG_FINAL};

/* -------------------------------------------------------*/

ap_uint<PBS> SWAP48(ap_uint<PBS> x)
{
#pragma HLS INLINE
    ap_uint<PBS> y = 0;
    for (size_t i = 0; i<PBS/8; i++) {
    #pragma HLS UNROLL
        ap_uint<8> t = x >> (PBS/8-1-i)*8;
        y ^= ((ap_uint<PBS>) t) << (i*8);
    }
    return y;
}

/* -------------------------------------------------------*/

void hls_ciphercore (
    BDO_STREAM& bdo_bus,
    bool *msg_auth_valid,
    BDI_STREAM& bdi_bus,
    KEY_STREAM& key_bus,
    bool decrypt,
    volatile bool key_update
    )
{
    /* Top-Level Interface */
#pragma HLS DATA_PACK variable=bdo_bus
#pragma HLS DATA_PACK variable=bdi_bus
#pragma HLS DATA_PACK variable=key_bus
#pragma HLS INTERFACE port=bdo_bus axis
#pragma HLS INTERFACE port=bdi_bus axis
#pragma HLS INTERFACE port=key_bus axis
#pragma HLS INTERFACE ap_vld  port=msg_auth_valid
#if __SYNTHESIS__
#else
    size_t dbg_loop = 0;
#endif

    public_bus bdi;
    output_bus bdo;
    secret_bus key;

    const ap_uint<1> Simonz[64] = {
        1, 0, 1, 0, 1, 1, 1, 1,
        0, 1, 1, 1, 0, 0, 0, 0,
        0, 0, 1, 1, 0, 1, 0, 0,
        1, 0, 0, 1, 1, 0, 0, 0,
        1, 0, 1, 0, 0, 0, 0, 1,
        0, 0, 0, 1, 1, 1, 1, 1,
        1, 0, 0, 1, 0, 1, 1, 0,
        1, 1, 0, 0, 1, 1};

#pragma HLS RESOURCE variable=Simonz core=ROM_1P_1S

    fsms_t fsm = S_INIT;
    static ap_uint<PBS> key_reg0;
    static ap_uint<PBS> key_reg1;
    ap_uint<4> cnt = 0;
    ap_uint<PBS> sl, sr, stateR, bdi_r;
    ap_uint<MAX_LEN> msglen = 0;
    bool is_eoi_r;
    bool is_eot_r;
    bool is_full_r;
    ap_uint<PBS/8> vbytes_r;
    ap_uint<PBS/8> padloc_r;

    // Key initialization (if needed)
    if (key_update == 1) {
        key_bus.read(key);
        key_reg0 = SWAP48(key.data);
        key_bus.read(key);
        key_reg1 = SWAP48(key.data);
    }

    // Initialize here
    bdi_bus.read(bdi);
    sr = SWAP48(bdi.data);
    sl = 0;
    stateR = 0;
    bool empty = bdi.eoi;
    bool loop = 1;

    while (loop) {
        // Add this clause to prevent infinite loop during debugging
#if !__SYNTHESIS__
        if (dbg_loop > 100) return;
        else dbg_loop++;
#endif
#if DBG
        std::cout << "Current fsm = " << fsm << " - cnt = " << cnt << std::endl;
#endif
       
        /* ==================== */
        /* Preparing input data */
        /* ==================== */
        ap_uint<PBS> k0, k1, x, y;
        ap_uint<PBS> t1 = ~(0xffffffffffffffff << nn);

        /* ========= */
        /* MAIN OP   */
        /* ========= */
        #if DBG
            printf("StateR: %012llX\n", (unsigned long long) stateR);
            printf("Input : %012llX %02llX\n", (unsigned long long) sl, (unsigned long long) sr);
        #endif
        x = sl; y = sr;
        k0 = key_reg0;
        k1 = key_reg1;

        for (size_t i=0; i<52; i++) {
#if DBG_RND
            printf("DAT(% 2d): %012llX %02llX\n", i, (unsigned long long) x, (unsigned long long) y);
            printf("Key(% 2d): %012llX %02llX\n", i, (unsigned long long) k0, (unsigned long long) k1);
#endif
            // Round
            ap_uint<PBS> tmp;
            tmp = x;
            x = y ^ ROTL(x, 1) & ROTL(x, 8) ^ ROTL(x, 2) ^ k0;
            y = tmp;
            // Calculate new round key
            ap_int<PBS> ktmp = k0;
#if 0
            // Note: Doesn't work for some reason.
            tmp = ROTL(k1, nn-3); 
            tmp = tmp ^ ROTL(tmp, nn-1);
#else
            tmp = ((k1 << 45) | (k1 >> 3));
            tmp ^= ((tmp << 47) | (tmp >> 1));
#endif
            k0 = k1;
            k1 = ~ktmp ^ tmp ^ ((ap_uint<2>) Simonz[i] ^ 3);
        }
        #if DBG
            printf("Output: %012llX %02llX\n", (unsigned long long) x, (unsigned long long) y);
        #endif
        // Input data
        /* Default values */
        bool is_eoi = 0;
        bool is_eot = 0;
        bool is_full = 0;
        bool is_data = 0;
        ap_uint<PBS> bdi_data;
        ap_uint<PBS> bditmp = 0;
        ap_uint<PBS/8> vbytes = 0;
        ap_uint<PBS/8> padloc = 0;
        ap_uint<PBS/8> vtmp = 0;
        ap_uint<PBS/8> ptmp = 0;


        if (fsm == S_AD || fsm == S_MSG) {
            bdi_bus.read(bdi); 

            is_data = (bdi.type == TYPE_D) ? 1:0;
            is_eoi = bdi.eoi;
            is_eot = bdi.eot;
            is_full = (bdi.size == 6) ? 1:0;
            // Swap endian of vbytes and padloc here
            ap_uint<PBS/8> tmp = bdi.valid_bytes;
            for (size_t i=0; i<PBS/8; i++) {
            #pragma HLS UNROLL
                ap_uint<1> bit = tmp >> (PBS/8-1-i);
                vtmp ^= (ap_uint<PBS/8>) bit << i; 
            }
            tmp = bdi.pad_loc;
            for (size_t i=0; i<PBS/8; i++) {
            #pragma HLS UNROLL
                ap_uint<1> bit = tmp >> (PBS/8-1-i);
                ptmp ^= (ap_uint<PBS/8>) bit << i; 
            }
            #if DBG
               std::cout << "Reading ..." << std::endl;
            #endif
            bditmp = SWAP48(bdi.data);
            if (fsm == S_AD && is_data) {
                bdi_data = 0x80;
                vbytes = 1;
                padloc = 0;
            } else {
                bdi_data = bditmp;
                vbytes = vtmp;
                padloc = ptmp;
            }
        } else if (fsm == S_STORE_MSG) {
            is_eoi = is_eoi_r;
            is_eot = is_eot_r;
            is_full = is_full_r;
            is_data = 1;
            bdi_data = bdi_r;
            vbytes = vbytes_r;
            padloc = padloc_r;
        } else if (fsm == S_AD_FULL || fsm == S_MSG_FULL) {
            bdi_data = 0x80;
            vbytes = 1;
            padloc = 0;
        } else {
            bdi_data = 0;
        }

        // Output data
        ap_uint<PBS> ct = y ^ stateR ^ bdi_data;
        ap_uint<PBS> data;
        ap_uint<PBS> pt = 0;
        for (size_t i=0; i<PBS/8; i++) {
        #pragma HLS UNROLL
            ap_uint<8> ptmp;
            if ((padloc >> i) & 1)
                ptmp = 0x80;
            else if ((vbytes >> i) & 1)
                ptmp = ct >> i*8;
            else 
                ptmp = 0;
            pt ^= ((ap_uint<PBS>) ptmp) << i*8; 
        }

        if ((fsm == S_MSG || fsm == S_STORE_MSG) && decrypt) {
            data = pt;
        } else {
            data = bdi_data;
        }

        bdi_r = bditmp;
        is_eoi_r = is_eoi;
        is_eot_r = is_eot;
        is_full_r = is_full;
        vbytes_r = vtmp;
        padloc_r = ptmp;

#if DBG
        printf("bdi  : %012llX\n", (unsigned long long) bdi_data);
        printf("ct   : %012llX\n", (unsigned long long) ct);
        printf("pt   : %012llX\n", (unsigned long long) pt);
#endif

        if (fsm == S_MSG || fsm == S_STORE_MSG) {
            bdo.data = SWAP48(ct);
            bdo_bus.write(bdo);
        }

        ap_uint<3> code;
        if (fsm == S_INIT) {
            code = 0x5;
        } else if (fsm == S_AD || fsm == S_AD_FULL) {
            code = 0x1;
        } else if (fsm == S_TAG) { // tag
            code = 0x3;
        } else {
            code = 0;
        }
        // sl = x; sr = y;
        sr = y ^ code ^ stateR;
        sl = x ^ data;
        stateR ^= x ^ data;

        /* ========= */
        /* Post processing  */
        /* ========= */

        switch (fsm) {
        case S_INIT:
            fsm = S_AD;
            if (empty) fsm = S_AD_FULL;
            else fsm = S_AD;
            #if DBG
                printf(" ==== AD\n");
            #endif
            break;
        case S_AD:
            if (is_data) {
                fsm = S_STORE_MSG;
                #if DBG
                    printf(" ==== MSG\n");
                #endif
            } else if (is_eot) {
                if (is_full) {
                    empty = is_eoi;
                    fsm = S_AD_FULL;
                } else {
                    if (is_eoi) {
                        fsm = S_MSG_FULL;
                    } else {
                        fsm = S_MSG;
                    }
                    #if DBG
                        printf(" ==== MSG\n");
                    #endif
                }
            } 
            break;
        case S_AD_FULL:
            if (empty) fsm = S_MSG_FULL;
            else fsm = S_MSG;
            #if DBG
                printf(" ==== MSG\n");
            #endif
            break;
        case S_STORE_MSG:
            if (is_eot) {
                if (is_full) fsm = S_MSG_FULL;
                else fsm = S_TAG;
            } else { fsm = S_MSG; }
            #if DBG
            if (is_eot && !is_full)
                printf(" ==== Tag\n");
            #endif
            break;
        case S_MSG:
            if (is_eot) {
                if (is_full) fsm = S_MSG_FULL;
                else fsm = S_TAG;
            }
            #if DBG
            if (is_eot && !is_full)
                printf(" ==== Tag\n");
            #endif
            break;
        case S_MSG_FULL:
            #if DBG
                printf(" ==== Tag\n");
            #endif
            fsm = S_TAG;
            break;
        case S_TAG:
            fsm = S_TAG_FINAL;
            break;
        default: // S_TAG_FINAL
            loop = 0;
        }
    }
    ap_uint<PBS> tag = sl ^ sr;
    tag = SWAP48(tag);
    if (decrypt) {
        bdi_bus.read(bdi); 
        *msg_auth_valid = (bdi.data == tag) ? 1:0;
        #if DBG
            std::cout << "Reading tag ..." << std::endl;
        #endif
    } else {
        bdo.data = tag;
        bdo_bus.write(bdo);
        #if DBG
            std::cout << "Writing tag ..." << std::endl;
        #endif
    }
    return;
}
