-------------------------------------------------------------------------------
--! @file       CipherCore.vhd
--! @author     Ekawat (ice) Homsirikamol
--! @brief      Top-level CipherCore for SIMONJAMBU - HLS
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.ALL;
use ieee.numeric_std.all;
use work.AEAD_pkg.all;

entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer :=48;   --! Data
        G_KEY_SIZE      : integer :=48;     --! Key
        G_TAG_SIZE      : integer :=48;    --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES      : integer := 3;
        --! Maximum supported AD/message/ciphertext length = 2^G_MAX_LEN-1
        G_MAX_LEN       : integer := SINGLE_PASS_MAX
    );
    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE       -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );
end entity CipherCore;

architecture structure of CipherCore is
    constant ZEROS          : std_logic_vector(7 downto 0) := (others => '0');
    constant BDI_BUS_LEN    : integer
                                := G_DBLK_SIZE+G_DBLK_SIZE/8*2+G_LBS_BYTES+7;
    constant BDO_BUS_LEN    : integer
                                := G_DBLK_SIZE+G_LBS_BYTES+1;
    constant BDI_PAD_LEN    : integer := (8 - (BDI_BUS_LEN mod 8)) mod 8;
    constant BDO_PAD_LEN    : integer := (8 - (BDO_BUS_LEN mod 8)) mod 8;
    constant WORDSIZE       : integer := 8;         --! Word size of data type used in HLS

    signal decrypt_r        : std_logic;
    signal start            : std_logic;
    signal ap_start_r       : std_logic;
    signal ap_done          : std_logic;
    signal ap_idle          : std_logic;
    signal ap_ready         : std_logic;
    signal rst_n            : std_logic;
    signal bdi_ready_s      : std_logic;

    signal bdi_bus          : std_logic_vector(BDI_PAD_LEN+G_DBLK_SIZE+G_DBLK_SIZE/8*2+G_LBS_BYTES+7 -1 downto 0);
    signal bdo_bus          : std_logic_vector(BDO_PAD_LEN+G_DBLK_SIZE+G_LBS_BYTES+1 -1 downto 0);
begin
    process(clk)
    begin
        if rising_edge(clk) then
            if rst = '1' then
                ap_start_r <= '0';
            else
                ap_start_r <= start;
                if (start = '1') then
                    decrypt_r <= decrypt;
                end if;
            end if;
        end if;
    end process;
    start <= (bdi_valid and not bdi_ready_s) and (ap_idle or ap_ready);
    bdi_ready <= bdi_ready_s;

    bdi_bus  <= ZEROS(BDI_PAD_LEN-1 downto 0) & bdi_partial & bdi_eot & bdi_eoi
                & bdi_type & bdi_size & bdi_pad_loc & bdi_valid_bytes & bdi;

    bdo_size <= bdo_bus(G_DBLK_SIZE+G_LBS_BYTES+1-1 downto G_DBLK_SIZE);
    bdo      <= bdo_bus(G_DBLK_SIZE-1 downto 0);

    rst_n <= not rst;
    u_crypto:
    entity work.hls_ciphercore(behav)
    port map (
        --! Global
        ap_clk                     => clk,
        ap_rst_n                   => rst_n,

        --! Start signals
        ap_start                   => ap_start_r,
        ap_done                    => ap_done,  --! Unused
        ap_idle                    => ap_idle,
        ap_ready                   => ap_ready,

        --! Output
        bdo_bus_V_TDATA            => bdo_bus,
        bdo_bus_V_TVALID           => bdo_valid,
        bdo_bus_V_TREADY           => bdo_ready,
        msg_auth_valid             => msg_auth_valid,
        msg_auth_valid_ap_vld      => msg_auth_done,

        --! Input
        bdi_bus_V_TDATA            => bdi_bus,
        bdi_bus_V_TVALID           => bdi_valid,
        bdi_bus_V_TREADY           => bdi_ready_s,
        key_bus_V_data_V_TDATA     => key,
        key_bus_V_data_V_TVALID    => key_valid,
        key_bus_V_data_V_TREADY    => key_ready,

        --! Info
        decrypt                    => decrypt_r,
        key_update                 => key_update
    );

end structure;